#  Copyright (c) 2023 Macrometa Corp All rights reserved.
import argparse
import json
import os
import time
from importlib import import_module
from inspect import ismethod

import pkg_resources
import psycopg2
from c8connector import C8Connector, ValidationException
from fastapi import FastAPI, Body, status, HTTPException
from psycopg2 import OperationalError
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT
from psycopg2.pool import SimpleConnectionPool
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
import uvicorn

package_filters = ['macrometa-source-', 'macrometa_source_', 'macrometa-target-', 'macrometa_target_',
                   'macrometa-integration-', 'macrometa_integration_']


class DB:
    def __init__(
            self, user: str, password: str, host: str, port: str, database: str = 'c8cws'
    ) -> None:
        self.user = user
        self.password = password
        self.host = host
        self.port = port
        self.database = database
        self.try_create_tables(database, host, password, port, user, 20)
        self.pool = SimpleConnectionPool(
            1, 25, user=user,
            password=password,
            host=host,
            port=port,
            database=database)

    @staticmethod
    def try_create_tables(database, host, password, port, user, max_tries):
        cursor, con = None, None
        backoff = 2
        done = False
        for i in range(max_tries):
            backoff = (i + 1) * backoff
            try:
                con = psycopg2.connect(user=user, password=password, host=host, port=port, database=database)
                con.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
                cursor = con.cursor()
                cursor.execute("CREATE TABLE workflow (uuid TEXT PRIMARY KEY, federation TEXT,"
                               " tenant TEXT, fabric TEXT, bin_workflow BYTEA, state JSONB);")
                done = True
            except OperationalError as e:
                print(f"Unable to connect to the central cloud RDS. {str(e).capitalize().strip()}."
                      f" Retrying in {backoff} seconds ...")
                time.sleep(backoff)
                continue
            except Exception as e:
                print(f"Skipping tables creation. {str(e).capitalize().strip()}.")
                done = True
                break
            finally:
                if cursor is not None:
                    cursor.close()
                if con is not None:
                    con.close()
        if not done:
            raise RuntimeError(f"Unable to connect to the central cloud RDS. "
                               f"Please check RDS connectivity and configurations.")

    def close(
            self
    ) -> None:
        self.pool.closeall()

    def update_state(self, uuid: str, state: object):
        state_json = json.dumps(state)
        conn = self.pool.getconn()
        cursor = conn.cursor()
        query = """
            UPDATE workflow
            SET state = %s
            WHERE uuid = %s;
        """
        cursor.execute(query, (state_json, uuid))
        conn.commit()
        cursor.close()
        self.pool.putconn(conn)

    def get_state(self, uuid: str):
        conn = self.pool.getconn()
        cursor = conn.cursor()
        query = """
            SELECT state
            FROM workflow
            WHERE uuid = %s;
        """
        cursor.execute(query, (uuid,))
        result = cursor.fetchone()
        cursor.close()
        self.pool.putconn(conn)
        if result is not None and result[0] is not None:
            return result[0]
        else:
            return None


class StateFileHandler(FileSystemEventHandler):
    def on_modified(self, event):
        if event.src_path.endswith('state.json'):
            save_state()


uuid = os.getenv("WORKFLOW_UUID")
state_save_interval = os.getenv("STATE_SAVE_INTERVAL", 60)
state_file_dir = os.getenv("STATE_DIR", "/project/eltworkflow/state")

db = DB(
    user=os.getenv("PG_USER") or os.getenv("RDS_USER"),
    password=os.getenv("PG_PASSWORD") or os.getenv("RDS_PASSWORD"),
    host=os.getenv("PG_HOST") or os.getenv("RDS_HOST"),
    port=os.getenv("PG_PORT") or os.getenv("RDS_PORT"),
    database=os.getenv("PG_DATABASE") or os.getenv("RDS_DATABASE", "c8cws")
)


def load_state():
    # Read from the db and write the state into state.json file.
    print(f"Loading the state of workflow {uuid} from the database.")
    try:
        state_content = db.get_state(uuid)
        if state_content:
            with open(f"{state_file_dir}/state.json", "w") as state_file:
                json.dump(state_content, state_file)
    except Exception as e:
        print(f"Couldn't load state of {uuid} from database. {e}")
        return


def save_state():
    # Read the state.json file and write the state to the DB.
    print(f"Saving the state of workflow {uuid} to the database.")
    try:
        with open(f"{state_file_dir}/state.json", "r") as state_file:
            state_content = json.load(state_file)
            if state_content:
                db.update_state(uuid, state_content)
    except Exception as e:
        print(f"Couldn't persist state of {uuid} to database. {e}")
        return


def start_state_observer():
    event_handler = StateFileHandler()
    observer = Observer()
    observer.schedule(event_handler, state_file_dir, recursive=False)
    print(f"Starting state observer on `{state_file_dir}/state.json`")
    observer.start()
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
    observer.join()


def expose_connector():
    # Create an argument parser with an argument for the port.
    parser = argparse.ArgumentParser(description="Expose the connector on the specified port.")
    parser.add_argument("port", type=int, help="Port number to expose the connector")
    args = parser.parse_args()
    port = args.port

    # Load all the installed packages in the current environment.
    installed_packages = pkg_resources.working_set
    for pkg in installed_packages:
        k = pkg.key
        if any(pf in k for pf in package_filters):
            try:
                # replace `-` of module name with `_`.
                k = k.replace("-", "_")
                import_module(k)
            except:
                pass

    # Discover the connector class that extends C8Connector.
    connector_classes = [cls for cls in C8Connector.__subclasses__() if issubclass(cls, C8Connector)]
    if len(connector_classes) != 1:
        raise Exception("Expected exactly one connector class, but found multiple or none.")

    # Create an instance of the connector class.
    connector_class = connector_classes[0]
    connector_instance = connector_class()

    # Create a FastAPI application.
    app = FastAPI(
        title=f"{connector_instance.name()} {connector_instance.type().title()} API",
        version=connector_instance.version()
    )

    # Expose methods as API endpoints.
    for method_name in dir(C8Connector):
        method = getattr(connector_instance, method_name)
        if ismethod(method) and method_name != "expose_connector":
            # Determine the HTTP method based on the method name
            http_method = "POST" if method_name in ("validate", "samples", "schemas") else "GET"

            # Generate a route path based on the method name
            route_path = f"/{method_name}"

            # Define a closure function to capture the method
            def create_route_handler(handler_method):
                if http_method == "POST":
                    async def route_handler(body: dict = Body(None)):
                        try:
                            return handler_method(body)
                        except ValidationException as e:
                            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=str(e))
                        except Exception as e:
                            raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=str(e))
                else:
                    async def route_handler():
                        try:
                            return handler_method()
                        except ValidationException as e:
                            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=str(e))
                        except Exception as e:
                            raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=str(e))
                return route_handler

            # Create the API route and bind it to the closure function
            if http_method == "POST":
                app.post(route_path)(create_route_handler(method))
            else:
                app.get(route_path)(create_route_handler(method))

    # Run the FastAPI application
    uvicorn.run(app, host="0.0.0.0", port=port)



#  Copyright (c) 2022 Macrometa Corp All rights reserved.
from enum import Enum
from datetime import datetime, timezone


class ConfigAttributeType(Enum):
    """C8Connector ConfigAttributeType"""
    STRING = "string"
    INT = "integer"
    BOOLEAN = "boolean"
    DATE = "date_iso8601"
    EMAIL = "email"
    PASSWORD = "password"
    OAUTH = "oauth"
    OPTIONS = "options"
    FILE = "file"
    ARRAY = "array"
    OBJECT = "object"
    HIDDEN = "hidden"


class SchemaAttributeType(Enum):
    """C8Connector SchemaAttributeType"""
    BOOLEAN = "bool"
    INT = "int"
    LONG = "long"
    FLOAT = "float"
    DOUBLE = "double"
    STRING = "string"
    OBJECT = "object"


class ConfigProperty:
    """C8Connector config property"""

    def __init__(self, name: str, display_name: str, type: ConfigAttributeType, is_mandatory: bool,
                 is_dynamic: bool, description: str, default_value: str = '', placeholder_value: str = ''):
        """
        Parameters
        :param name: Name of the config property
        :param display_name: Display name of the config property
        :param type: Attribute type of the config property
        :param is_mandatory: Boolean indicating whether the property is mandatory
        :param is_dynamic: Boolean indicating whether the property is dynamic
        :param description: Description of the config property
        :param default_value: Default value of the property
        :param placeholder_value: Placeholder value of the property
        """
        self.name = name
        self.type = type
        self.display_name = display_name
        self.is_mandatory = is_mandatory
        self.is_dynamic = is_dynamic
        self.description = description
        self.default_value = default_value
        self.placeholder_value = placeholder_value


class SchemaAttribute:
    """C8Connector Attribute"""

    def __init__(self, name: str, type: SchemaAttributeType):
        self.name = name
        self.type = type


class Schema:
    """C8Connector Schema"""

    def __init__(self, name: str, attributes: list[SchemaAttribute]):
        self.name = name
        self.attributes = attributes


class Sample:
    """C8Connector Sample"""

    def __init__(self, schema: Schema, data: list[dict]):
        self.schema = schema
        self.data = data


class ValidationException(Exception):
    """C8Connector Validation Exception"""


class C8ConnectorMeta(type):
    """C8Connector metaclass"""

    def __instancecheck__(cls, instance):
        return cls.__subclasscheck__(type(instance))

    def __subclasscheck__(cls, subclass):
        return (
                hasattr(subclass, 'name') and callable(subclass.name) and
                hasattr(subclass, 'package_name') and callable(subclass.package_name) and
                hasattr(subclass, 'version') and callable(subclass.version) and
                hasattr(subclass, 'type') and callable(subclass.type) and
                hasattr(subclass, 'description') and callable(subclass.description) and
                hasattr(subclass, 'validate') and callable(subclass.validate) and
                hasattr(subclass, 'samples') and callable(subclass.samples) and
                hasattr(subclass, 'schemas') and callable(subclass.schemas) and
                hasattr(subclass, 'reserved_keys') and callable(subclass.reserved_keys) and
                hasattr(subclass, 'config') and callable(subclass.config) and
                hasattr(subclass, 'capabilities') and callable(subclass.capabilities)
        )


class C8Connector(metaclass=C8ConnectorMeta):
    """C8Connector superclass"""

    def name(self) -> str:
        """Returns the name of the connector."""
        pass

    def package_name(self) -> str:
        """Returns the package name of the connector (i.e. PyPi package name)."""
        pass

    def version(self) -> str:
        """Returns the version of the connector."""
        pass

    def type(self) -> str:
        """Returns the type of the connector."""
        pass

    def description(self) -> str:
        """Returns the description of the connector."""
        pass

    def validate(self, integration: dict) -> None:
        """Validate given configurations against the connector.
        If invalid, throw a ValidationException with the cause.
        """
        pass

    def samples(self, integration: dict) -> list[Sample]:
        """Fetch sample data using the given configurations.
        For Source - Return the first 10 samples (data and corresponding schema).
        If _key, _id or _rev exists in source data/schema then append "_" to the key as long as the new key doesn't already exist in the source data.
        If _key is also the primary key of the source data then do not append "_" to "_key".
        If the configurations are not valid throw ValidationException with the cause.

        For Target - Return an empty list.
        """
        pass

    def schemas(self, integration: dict) -> list[Schema]:
        """Get supported schemas using the given configurations.
        For Source - Return the schema of the source data.
        If _key, _id or _rev exists in source data/schema then append "_" to the key as long as the new key doesn't already exist in the source data/schema.
        If _key is also the primary key of the source data then do not append "_" to "_key".
        If the configuration is not valid throw ValidationException with the cause.

        For Target - Return an empty list.
        """
        pass

    def reserved_keys(self) -> list[str]:
        """List of reserved keys for the connector.
        For Source - It should always be empty list [].

        For Target - It should contain all the reserved keys for the target DB as a list of string (if reserved keys exists), else an empty list [].
        It can be keywords which are always auto generated, fixed, internal or a fixed key as primary key for the target DB.
        If there is a fixed primary key it should always be specified as the first element of the list.
        Else if there isn't a fixed primary key but there are other reserved keys then the first element should be an empty string followed by the
        list of reserved keys, Example: ["", "reservedkey1", "reservedkey2"].
        """
        pass

    def config(self) -> list[ConfigProperty]:
        """Get configuration parameters for the connector."""
        pass

    def capabilities(self) -> list[str]:
        """Return the capabilities[1] of the connector.
        [1] https://docs.meltano.com/contribute/plugins#how-to-test-a-tap
        """
        pass


def ensure_datetime(time_extracted):
    if isinstance(time_extracted, str):
        time_extracted = datetime.strptime(time_extracted, "%Y-%m-%dT%H:%M:%S.%fZ")
        # Make the datetime object timezone-aware by setting it to UTC timezone
        time_extracted = time_extracted.replace(tzinfo=timezone.utc)
    elif time_extracted.tzinfo is None:
        # If the datetime object is timezone-naive, set it to UTC timezone
        time_extracted = time_extracted.replace(tzinfo=timezone.utc)
    return time_extracted



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'c8connector'}

packages = \
['c8connector', 'c8connector.util', 'util']

package_data = \
{'': ['*']}

install_requires = \
['fastapi==0.65.1',
 'psycopg2-binary==2.9.3',
 'uvicorn==0.18.0',
 'watchdog==3.0.0']

entry_points = \
{'console_scripts': ['expose-connector = c8connector.util:expose_connector',
                     'load-workflow-state = c8connector.util:load_state',
                     'start-state-observer = '
                     'c8connector.util:start_state_observer']}

setup_kwargs = {
    'name': 'c8connector',
    'version': '0.0.29',
    'description': 'C8 Connector Interface',
    'long_description': '# Implementing C8 Connectors.\n\nUsers can extend `C8Connector` interface and develop 3 types of connectors.\n\n1. Source Connectors (Connectors that ingest data)\n2. Target Connectors (Connectors that export data)\n3. Integration Connectors (Generic integrations for other services)\n\nWhen developing these connectors, developers must adhere to a few guidelines mentioned below.\n\n## Naming the Connector\n\n- Package name of the connector must be in the `macrometa-{type}-{connector}` format (i.e `macrometa-source-postgres`).\n- Module name of the connector must be in the `macrometa_{type}_{connector}` format (i.e `macrometa_source_postgres`).\n\n## Project structure (package names and structure)\n\n- Project source code must follow the below structure.\n```text\n.\n├── LICENSE\n├── README.md\n├── GETTING_STARTED.md\n├── macrometa_{type}_{connector}\n│        ├── __init__.py\n│        └── main.py\n│        └── {other source files or modules}\n├── pyproject.toml\n└── setup.cfg\n```\n- Within the `/macrometa_{type}_{connector}/__init__.py` there must be a class which implements `C8Connector` interface.\n\n## Dependencies/Libraries and their versions to use.\n\n- Connectors must only use following dependencies/libraries and mentioned versions\' when developing.\n```text\npython = ">=3.7"\nc8connector = "latest"\npipelinewise-singer-python = "1.2.0"\n```\n- Developers must not use `singer-sdk` or any other singer sdk variants other than `pipelinewise-singer-python`.\n\n## Connector specific documentation\n\n- Every connector project should have a GETTING_STARTED.md file, documenting the connector configuration and all other requirements for the connector.\n  It should be formatted like a User-Facing document and it should also provide the necessary instructions for the end user to be able to use the connector.\n  \n  Developers can follow the Generic Template available [here](https://github.com/Macrometacorp/c8connector/blob/main/GETTING_STARTED.md) and apply any necessary changes required on top of it for the specific connector.\n\n## Resolving reserved key conflicts between macrometa collection and external DB\n\n- For Source connectors:\n  Macrometa collection(document) has the following reserved keys, `_key`, `_id` and `_rev`. `_key` is the primary key and hence, it will have the value of the primary key of source data and `_id` and `_rev` are always autogenerated. So if `_key`, `_id`, `_rev` also exists in source data (assuming _key is not the primary key of source data) then these values from source data would be lost.\n  Hence we need to append an additional `_` to these reserved keys if they are present in source data. If `_key` is the primary key of the source data itself then no need to append `_` to `_key`. We should also check that the new key generated doesn\'t exist in the source data, If it exists then keep appending `_`.\n  \n  During the actual workflow run, this logic is implemented at the target level (macrometa-target-collection). But we also need the same to be implemented at source connector levels for `samples` and `schemas` API.\n  Refer [PR] (https://github.com/Macrometacorp/macrometa-source-postgres/)\n\n\n- For Target connectors:\n  As seen for source connector with target as (macrometa-target-collection), the same reserved keys conflict can arise in case of target connectors too, where External database might have some fixed reserved keys which might be their primary key, autogenerated or internal key. So if such reserved keys also exist in source collection then these values from source collection will be lost in the target data.\n  \n  In such cases we should first specify all the reserved keys as a list of string in the `reserved_keys` field of target connector. If there is a fixed primary key it should always be specified as the first element of the list, else if there isn\'t a fixed primary key but there are other reserved keys then the first element should be an empty string followed by the list of reserved keys, Example: ["", "reservedkey1", "reservedkey2"]. If no reserved keys exist return an empty list []. Refer [PR] (https://github.com/Macrometacorp/macrometa-target-collection/pull/9)\n\n  In addition to this, we also need to implement the logic of appending `_` to the reserved keys (only if they exist in source collection) before writing the data in the external DB at the target connector level. If `_key` is also the reserved primary key of the target external DB then no need to append `_` to this reserved primary key. We should also check that the new key generated doesn\'t exist in the source collection, If it exists then keep appending `_`. Refer [PR] (https://github.com/Macrometacorp/macrometa-target-collection/pull/10)\n\n  > **_NOTE:_** This is applicable only when there are certain keys reserved in the external database.\n\n\n## Adding metrics to the connectors\n- For Source connectors:\n  We support the following ingest metrics for source connectors:\n    ingested_bytes, ingested_documents, ingest_errors, ingest_lag\n  Out of the above 4 metrics, we only need to increment ingest_errors at the source connector level whenever there is an error and start the prometheus client http server at port 8000.\n  The rest of them are calculated at macrometa target collection level.\n  But to calculate ingest_lag metrics the source connector needs to send the current timestamp in `time_extracted` property of the Singer Record Message.\n  You can refer to other source connectors, for example:\n  https://github.com/Macrometacorp/macrometa-source-postgres/pull/14/files\n\n- For Target connectors:\n  We support the following ingest metrics for target connectors:\n    exported_bytes, exported_documents, export_errors, export_lag\n  Out of the above 4 metrics, exported_documents and exported_bytes are calculated at macrometa source collection level.\n  We need to increment export_errors at the target connector level whenever there is an error and start the prometheus client http server at port 8001 and we also need to calculate export_lag.\n  `export_lag` is nothing but the time difference in seconds between the `time_extracted` property in Singer Record Message, which is sent by macrometa source collection connector, and the current timestamp (UTC timezone).\n  You can refer to other target connectors, for example:\n  https://github.com/Macrometacorp/macrometa-target-postgres/pull/31/files\n\n\n## Samples\n- Postgres Source Connector: [Git Repository](https://github.com/Macrometacorp/macrometa-source-postgres)\n- Oracle Source Connector: [Git Repository](https://github.com/Macrometacorp/macrometa-source-oracle)\n- C8 Collections target Connector: [Git Repository](https://github.com/Macrometacorp/macrometa-target-collection)\n',
    'author': 'Macrometa',
    'author_email': 'info@macrometa.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8.1,<3.11',
}


setup(**setup_kwargs)
